@@ -225,7 +225,6 @@ class DefectEqSolver:
             opt_params_in["opt_Nsites_output"]="both"
         if not "opt_Nsites_output_Edef_atlast" in opt_params_in.keys():
             opt_params_in["opt_Nsites_output_Edef_atlast"]=True
-
         if opt_params_in["opt_Nsites_output"]=="both":
             bool_output_std=True
             bool_output_file=True
@@ -314,8 +313,6 @@ class DefectEqSolver:
         for e1,nat1 in Natoms_def.items():
             if not e1 in elems_list:
                 elems_list.append(e1)
-        # for e1,cp1 in chempots_in.items():
-        # for e1,nat1 in Natoms_def.items():
         for e1 in elems_list:
             cp1=chempots_in[e1]
             outdata_all.append(["","","","","chempot_"+e1,cp1])
@@ -358,40 +355,108 @@ class DefectEqSolver:
             outdata_all.append(outdata)
         return outdata_all
 
-    def auxopt_next_step(self,x0,x1,y0,y1,delx):
-        if x0<x1: 
-            if np.fabs(y0)>np.fabs(y1):
-                nextx=x1+delx
-            else:
-                nextx=x0-delx
-        else:
-            if np.fabs(y0)>np.fabs(y1):
-                nextx=x1-delx
-            else:
-                nextx=x0+delx
-        return nextx
-
-    def opt_chempot_1dim(self,func,x0,x1,delx_step=0.1,fout=None,tol=1e-7,maxiter=1000):
-        y0=func(x0)
-        x1=x0+delx_step
+    def opt_steady_line_minimization(self,func_resid,vect_grad,cp0,step_size,fout=None,tol_in=1e-7,maxiter=1000):
+        fout.write(f" ------ Starting step-wize minimization with step size of {step_size} -----\n")
+        residArray0=func_resid(cp0)
+        resid0=np.sqrt(np.square(residArray0).mean(axis=0))
+        vect_ms=0.0
+        for i1,v1 in enumerate(vect_grad):
+            vect_ms+=v1**2
+        vect_ms=vect_ms**0.5
+        vect1=[]
+        for i1,v1 in enumerate(vect_grad):
+            vect1.append(v1/vect_ms)
+        cp1=[]
+        for i1,v1 in enumerate(vect1):
+            cp1.append(cp0[i1]+step_size*v1)
         i_loop=1
         while i_loop<maxiter:
-            y1=func(x1)
-            if y0*y1<0.0:
+            residArray1=func_resid(cp1)
+            resid1=np.sqrt(np.square(residArray1).mean(axis=0))
+            if resid1>resid0:
                 break
             else:
-                x2=self.auxopt_next_step(x0,x1,y0,y1,delx_step)
-                x0=x1
-                y0=y1
-                x1=x2
+                cp2=[]
+                for i1,v1 in enumerate(vect1):
+                    cp2.append(cp1[i1]+step_size*v1)
+                cp0=cp1
+                residArray0=residArray1
+                resid0=resid1
+                cp1=cp2
             i_loop+=1
-        fout.write(" ------ Starting minimization -----\n")
+        print("finished: step-wize minimization")
+        fout.write(" ------ Starting golden minimization -----\n")
         print(" [opt_Natoms] Starting minimization")
-        res_x=optimize.brenth(func,x0,x1,xtol=1e-8,rtol=tol)
-        # x_res=optimize.ridder(func,x0,x1,xtol=1e-8,rtol=tol)
-        # x_res=optimize.root_scalar(func,x0=(x0+x1)*0.5,x1=x0,xtol=1e-8,rtol=tol,method="secant")
-        res_y=func(res_x)
-        return ([res_x],[res_y])
+        def func_resid_local(x_in):
+            cp2=[]
+            for i1 in range(len(cp0)):
+                cp2.append(x_in*cp0[i1]+(1.0-x_in)*cp1[i1])
+            residArray2=func_resid(cp2)
+            resid2=np.sqrt(np.square(residArray2).mean(axis=0))
+            return resid2
+        res3=optimize.minimize_scalar(func_resid_local,bracket=(0.0,1.0),method="golden",tol=tol_in,options={'xtol':tol_in})
+        cp3=[]
+        for i1 in range(len(cp0)):
+            cp3.append(res3.x*cp0[i1]+(1.0-res3.x)*cp1[i1])
+        residArray3=func_resid(cp3)
+        resid3=np.sqrt(np.square(residArray3).mean(axis=0))
+        print("finished: opt_steady_line_minimization")
+        return(cp3,residArray3)
+
+    def opt_chempot_multi(self,func_resid,func_grad,cp0,delta_cp=0.1,fout=None,tol=1e-6,maxiter=1000):
+        cp2=copy.copy(cp0)
+        delta_cp2=delta_cp
+        residArray2=func_resid(cp2)
+        vgrad2=func_grad(residArray2)
+        resid2=np.sqrt(np.square(residArray2).mean(axis=0))
+        CGalgo="PR" ### PR may be best
+        for iloop1 in range(maxiter):
+            ###### CG
+            if iloop1==0:
+                cgvect2=copy.copy(vgrad2)
+            else:
+                cgvect1=copy.copy(cgvect2)
+                cgfact1=0.0
+                cgfact2=0.0
+                ## Fletcher-Reeves
+                if CGalgo=="FR":
+                    for i1 in range(len(cp0)):
+                        cgfact2+=vgrad2[i1]**2
+                        cgfact1+=vgrad1[i1]**2
+                ## Polak-Ribiere
+                elif CGalgo=="PR":
+                    for i1 in range(len(cp0)):
+                        cgfact2+=vgrad2[i1]*(vgrad2[i1]-vgrad1[i1])
+                        cgfact1+=vgrad1[i1]**2
+                ## Hestenes-Stiefel
+                elif CGalgo=="HS":
+                    for i1 in range(len(cp0)):
+                        cgfact2+=vgrad2[i1]*(vgrad2[i1]-vgrad1[i1])
+                        cgfact1+=cgvect1[i1]*(vgrad2[i1]-vgrad1[i1])
+                ## Dai-Yuan
+                elif CGalgo=="DY":
+                    for i1 in range(len(cp0)):
+                        cgfact2+=vgrad2[i1]**2
+                        cgfact1+=cgvect1[i1]*(vgrad2[i1]-vgrad1[i1])
+                #######################
+                betaCG2=cgfact2/cgfact1
+                if CGalgo=="PR" and betaCG2<0.0:
+                    betaCG2=0.0
+                cgvect2=[]
+                for i1 in range(len(cp0)):
+                    cgvect2.append(vgrad2[i1]+betaCG2*cgvect1[i1])
+            vgrad1=copy.copy(vgrad2)
+            fout.write(f" ----- New search starting ({iloop1+1}) -----\n")
+            fout.write(f" ----- Gradient for the line search = {cgvect2} -----\n")
+            (cp3,residArray2)=self.opt_steady_line_minimization(func_resid,cgvect2,cp2,delta_cp2,fout,tol,maxiter)
+            delta_cp2=delta_cp/(iloop1+2.0)
+            cp2=cp3
+            vgrad2=func_grad(residArray2)
+            resid2=np.sqrt(np.square(residArray2).mean(axis=0))
+            if resid2<tol:
+                break
+        print("finished: opt_chempot_multi")
+        return(cp2,residArray2)
 
     def opt_Natoms(self,temperature_in,chempots_in,Natoms_target_in,
                    opt_params_in,root_outfiles,plot_params_in):
@@ -404,7 +469,10 @@ class DefectEqSolver:
         if not "opt_Natoms_output" in opt_params_in.keys():
             opt_params_in["opt_Natoms_output"]="both"
         if not "opt_Natoms_output_every_step" in opt_params_in.keys():
-            opt_params_in["opt_Natoms_output_every_step"]=True
+            opt_params_in["opt_Natoms_output_every_step"]=False
+        if not "opt_Natoms_delta_chempot" in opt_params_in.keys():
+            opt_params_in["opt_Natoms_delta_chempot"]=0.1
+
         self.reset_id_loop_Nsites()
         self.reset_id_loop_eFermi()
         Natoms_elem_list=[]
@@ -445,7 +513,8 @@ class DefectEqSolver:
             fout1.write(str1)
             str_outlabel=""
             for e1 in Natoms_elem_list:
-                str_outlabel+=f"   mu_{e1}           |   Nat_{e1}          |    delNat_{e1}       |"
+                str_outlabel+=f"   mu_{e1:<3}          |   Nat_{e1:<3}         |    delNat_{e1:<3}      |"
+            str_outlabel+=f"     residual       |"
             fout1.write(str_outlabel+"\n")
         if opt_params_in["opt_Natoms_output_every_step"]:
             opt_params_in["opt_Nsites_output_Edef_atlast"]=True
@@ -453,8 +522,10 @@ class DefectEqSolver:
             opt_params_in["opt_Nsites_output_Edef_atlast"]=False
         def output_local(cp_new,Nat_new,delNat_new):
             str1=""
+            resid_new=np.sqrt(np.square(delNat_new).mean(axis=0))
             for ie1,e1 in enumerate(Natoms_elem_list):
                 str1+=f" {cp_new[ie1]:17.14f} | {Nat_new[ie1]:17.14f} | {delNat_new[ie1]:18.11e} |"
+            str1+=f" {resid_new:18.11e} |"
             if bool_output_std:
                 print(" [Natoms] "+str_outlabel)
                 print(" [Natoms] "+str1)
@@ -462,40 +533,31 @@ class DefectEqSolver:
             if bool_output_file:
                 fout1.write(str1+"\n")
         print(" [opt_Natoms] Starating")
-        if len(Natoms_elem_list)==1:
-            def func_Natoms1_local(cp_in):
-                cp_list=copy.deepcopy(chempots_in)
-                cp_list[Natoms_elem_list[0]]=cp_in
-                self.opt_Nsites_for_chempots(temperature_in,cp_list,opt_params_in,root_outfiles,plot_params_in)
-                del_nat=0.0
-                nat_list=self.defects.get_Natoms_list()
-                nat2=nat_list[Natoms_elem_list[0]]
-                nat3=Natoms_target_list[0]
-                del_nat=float(nat2)-float(nat3)
-                output_local([cp_in],[nat2],[del_nat])
-                return del_nat
-            (Natoms_chempot_results_list,res_delNat)=self.opt_chempot_1dim(func_Natoms1_local,
-                    Natoms_chempot0_list[0],Natoms_chempot0_list[0]+1.0,0.1,
-                    fout1,opt_params_in["opt_Natoms_1dim_tol"])
-        else:
-        # if opt_params_in["opt_Natoms_type"]=="optimize":
-            print("  ERROR:: Not-yet supprted for opt_Natoms of multi-elements!")
-            sys.exit()
-            def opt_Nsite_chempot_local(x_in):
-                cp1=copy.deepcopy(chempots_in)
-                for ix1,x1 in enumerate(x_in):
-                    cp1[Natoms_elem_list[ix1]]=x1
-                self.opt_Nsites_for_chempots(temperature_in,cp1,opt_params_in,root_outfiles)
-                del_nat=0.0
-                nat_list=self.defects.get_Natoms_list()
-                for ix1,x1 in enumerate(x_in):
-                    nat2=nat_list[Natoms_elem_list[ix1]]
-                    nat3=Natoms_target_list[ix1]
-                    del_nat+=(nat2-nat3)**2
-                residual_Natoms=del_nat
-                return residual_Natoms
-            while residual_Natoms>opt_params_in["opt_Natoms_tol"]:
-                optimize.minimize(opt_Nsite_chempot_local,Natoms_chempot0_list,method="Nelder-Mead")
+        def func_resid_fixNatoms(cp_in):
+            cp1=copy.deepcopy(chempots_in)
+            for icp1,cp_in1 in enumerate(cp_in):
+                cp1[Natoms_elem_list[icp1]]=cp_in1
+            self.opt_Nsites_for_chempots(temperature_in,cp1,opt_params_in,root_outfiles)
+            del_nat_array=[]
+            nat_array=[]
+            nat_dict=self.defects.get_Natoms_list()
+            for icp1,cp_in1 in enumerate(cp_in):
+                nat2=nat_dict[Natoms_elem_list[icp1]]
+                nat3=Natoms_target_list[icp1]
+                del_nat_array.append(nat2-nat3)
+                nat_array.append(nat2)
+            output_local(cp_in,nat_array,del_nat_array)
+            return del_nat_array
+        def func_grads_fixNatoms(residArray_in):
+            jacob1=self.defects.get_Jacobian(Natoms_elem_list,temperature_in)
+            vect_grads=np.zeros(len(Natoms_elem_list))
+            for ie1,e1 in enumerate(Natoms_elem_list):
+                for ie2,e2 in enumerate(Natoms_elem_list):
+                    vect_grads[ie1]-=2.0*residArray_in[ie2]*jacob1[ie2][ie1]
+            return vect_grads
+        (Natoms_chempot_results_list,res_delNat)=self.opt_chempot_multi(
+                func_resid_fixNatoms,func_grads_fixNatoms,
+                Natoms_chempot0_list,opt_params_in["opt_Natoms_delta_chempot"],fout1,opt_params_in["opt_Natoms_tol"])
         print(" [opt_Natoms] Finished")
         cp_list=copy.deepcopy(chempots_in)
         for iat1,nat1 in enumerate(Natoms_elem_list):
@@ -515,9 +577,12 @@ class DefectEqSolver:
         self.defects=Defects(self.host,elements_in,input_paths_in)
         self.reset_id_loop_Nsites()
         self.reset_id_loop_eFermi()
+        self.chempots_fixNatoms="NONE"
         return
-    
+
     def opt_coordinator(self,eq_cond_in,solver_params_in,root_outfiles,plot_params_in):
+        if not "opt_Natoms_chempot_init" in solver_params_in.keys():
+            solver_params_in["opt_Natoms_chempot_init"]="fix"
         temp=eq_cond_in["T"]
         chempots={}
         for l1, v1 in eq_cond_in.items():
@@ -536,8 +601,28 @@ class DefectEqSolver:
             self.output_defEne_wrt_eFermi(temp,chempots,root_outfiles,plot_params_in)
             return "NONE"
 
-        if "fix_Natoms" in eq_cond_in.keys():
-            defect_densities=self.opt_Natoms(temp,chempots,eq_cond_in["fix_Natoms"],solver_params_in,root_outfiles,plot_params_in)
+        bool_fixNatoms=False
+        self.eqlabel_fixNatoms="NONE"
+        for k1 in eq_cond_in.keys():
+            if k1=="fix_Natoms":
+                bool_fixNatoms=True
+                self.eqlabel_fixNatoms="fix_Natoms"
+            elif k1=="fix_Natoms_linked":
+                bool_fixNatoms=True
+                self.eqlabel_fixNatoms="fix_Natoms_linked"
+        if bool_fixNatoms:
+            if solver_params_in["opt_Natoms_chempot_init"]=="update" and self.chempots_fixNatoms!="NONE":
+                for ie1,eq1 in enumerate(eq_cond_in[self.eqlabel_fixNatoms]):
+                    e1=eq1["element"]
+                    eq1["chempot_init"]=self.chempots_fixNatoms[e1]
+            defect_densities=self.opt_Natoms(temp,chempots,eq_cond_in[self.eqlabel_fixNatoms],solver_params_in,root_outfiles,plot_params_in)
+            if solver_params_in["opt_Natoms_chempot_init"]=="update":
+                self.chempots_fixNatoms={}
+                for ie1,eq1 in enumerate(eq_cond_in[self.eqlabel_fixNatoms]):
+                    e1=eq1["element"]
+                    for d1 in defect_densities:
+                        if d1[-2]=="chempot_"+e1:
+                            self.chempots_fixNatoms[e1]=float(d1[-1])
         else:
             defect_densities=self.opt_Nsites_for_chempots(temp,chempots,solver_params_in,root_outfiles,plot_params_in)
 
